from dataclasses import dataclass
from enum import IntEnum, Enum, unique

from utils import register_names

REG_DICT = { i: reg for i, reg in enumerate(register_names()) }

@unique
class Instruction(str, Enum):
    LUI = "LUI"
    AUIPC = "AUIPC"
    JAL = "JAL"
    JALR = "JALR"
    BEQ = "BEQ"
    BNE = "BNE"
    BLT = "BLT"
    BGE = "BGE"
    BLTU = "BLTU"
    BGEU = "BGEU"
    LB = "LB"
    LH = "LH"
    LW = "LW"
    LBU = "LBU"
    LHU = "LHU"
    SB = "SB"
    SH = "SH"
    SW = "SW"
    ADDI = "ADDI"
    SLTI = "SLTI"
    SLTIU = "SLTIU"
    XORI = "XORI"
    ORI = "ORI"
    ANDI = "ANDI"
    SLLI = "SLLI"
    SRLI = "SRLI"
    SRAI = "SRAI"
    ADD = "ADD"
    SUB = "SUB"
    SLL = "SLL"
    SLT = "SLT"
    SLTU = "SLTU"
    XOR = "XOR"
    SRL = "SRL"
    SRA = "SRA"
    OR = "OR"
    AND = "AND"
    FENCE = "FENCE"
    FENCE_I = "FENCE_I"
    ECALL = "ECALL"
    EBREAK = "EBREAK"
    CSRRW = "CSRRW"
    CSRRS = "CSRRS"
    CSRRC = "CSRRC"
    CSRRWI = "CSRRWI"
    CSRRSI = "CSRRSI"
    CSRRCI = "CSRRCI"


class DecodeError(Exception):
    pass


@dataclass
class RVInst:
    mnemonic: str
    opcode: int

    def __str__(self):
        # Ignore the opcode for now.
        return self.mnemonic


@dataclass
class RType(RVInst):
    rd: int
    funct3: int
    rs1: int
    rs2: int
    funct7: int

    def __str__(self):
        return super(RType,self).__str__() + f",rd:{hex(self.rd)},"\
            f"rs1:{hex(self.rs1)},rs2:{hex(self.rs2)}"


@dataclass
class IType(RVInst):
    imm: int
    rd: int
    funct3: int
    rs1: int

    def __str__(self):
        return super(IType, self).__str__() + \
            f",rd:{REG_DICT[self.rd]},rs1:{REG_DICT[self.rs1]}," \
            f"imm:{hex(self.imm)}"

@dataclass
class SType(RVInst):
    imm: int
    funct3: int
    rs1: int
    rs2: int

    def __str__(self):
        return super(SType, self).__str__() + \
            f",imm:{hex(self.imm)},funct3:{hex(self.funct3)}," \
            f"rs1:{REG_DICT[self.rs1]},rs2:{REG_DICT[self.rs2]}"

@dataclass
class BType(RVInst):
    imm: int
    funct3: int
    rs1: int
    rs2: int

    def __str__(self):
        return super(BType, self).__str__() + \
            f"rs1:{REG_DICT[self.rs1]},rs2:{REG_DICT[self.rs2]},imm:{hex(self.imm)}"
    
@dataclass
class UType(RVInst):
    imm: int
    rd: int

    def __str__(self):
        return super(UType, self).__str__() + \
            f",rd:{REG_DICT[self.rd]},imm:{hex(self.imm)}"

@dataclass
class JType(RVInst):
    imm: int
    rd: int

    def __str__(self):
        return super(JType, self).__str__() + \
            f",rd:{REG_DICT[self.rd]},imm:{hex(self.imm)}"

@unique
class Opcode(IntEnum):
    # These instructions are identified using only
    # the opcode field
    LUI      = 0b0110111
    AUIPC    = 0b0010111
    JAL      = 0b1101111
    JALR     = 0b1100111
    #
    # Branch instructions:
    # BEQ, BNE, BLT, BGE, BLTU, BGEU
    #
    BRANCH =  0b1100011
    #
    # Load instructions:
    # LB, LH, LW, LBU, LHU
    #
    LOAD = 0b0000011
    #
    # Store instructions:
    # SB, SH, SW
    STORE = 0b0100011
    #
    # Immediate instructions:
    # ADDI, SLTI, SLTIU, XORI, ORI, ANDI, SLLI, SRLI, SRAI
    #
    IMMEDIATE = 0b0010011
    #
    # Arithmetic/logical instructions:
    # ADD, SUB, SLL, SLT, SLTU, XOR, SRL, SRA, OR, AND
    #
    ARITHMETIC = 0b0110011
    #
    # Fence instructions:
    # FENCE, FENCE.I
    #
    FENCES = 0b0001111
    #
    # System instructions:
    # ECALL, EBREAK, CSRRW, CSRRS, CSRRC, CSRRWI, CSRRSI, CSRRCI
    #
    SYSTEM = 0b1110011


class Funct3(IntEnum):
    JALR = BEQ = LB = SB = ADDI = ADD = SUB = FENCE = ECALL = EBREAK = 0b000
    BNE = LH = SH = SLLI = SLL = FENCE_I = CSRRW = 0b001
    BLT = LBU = XORI = XOR = 0b100
    BGE = LHU = SRLI = SRAI = SRL = SRA = CSRRWI = 0b101
    BLTU = ORI = OR = CSRRSI = 0b110
    BGEU = ANDI = AND = CSRRCI = 0b111
    LW = SW = SLTI = SLT = CSRRS = 0b010
    SLTIU = SLTU = CSRRC = 0b011


class Funct7(IntEnum):
    SLLI = SRLI = ADD = SLL = SLT = SLTU = XOR = SRL = OR = AND = 0b0000000
    SRAI = SUB = SRA = 0b0100000


def jtype_imm(inst: int) -> int:
    """
    Decodes J-type immediates.

    Args:
        inst(int): Word-length instruction in bytes.
    Returns:
        int: J-type immediate value.
    """
    return sign_extend((((inst >> 31) & 0b1) << 20) |
                       (((inst >> 12) & 0b11111111) << 12) |
                       (((inst >> 20) & 0b1) << 11) |
                       (((inst >> 21) & 0x3FF) << 1), 20)

def btype_imm(inst: int, signed=True) -> int:
    """
    Decodes B-type immediates.

    Args:
        inst(int): Word-length instruction in bytes.
    Returns:
        int: B-type immediate value.
    """
    b_imm1 = (inst >> 7) & 0b11111
    b_imm2 = (inst >> 25) & 0b1111111
    x = ((b_imm2 >> 5) << 12) | \
        ((b_imm1 & 0b1) << 11) | \
        ((b_imm2 & 0b111111) << 5) | \
        ((b_imm1 >> 1) << 1)

    return sign_extend(x, 12) if signed else x
    
def itype_imm(inst: int, signed=True) -> int:
    """
    Decodes an I-type immediate value.

    Args:
       inst(int): Word-length instruction in bytes.

    Returns:
       int: I-type immediate value.
    """
    x = (inst >> 20) & 0xFFF
    return sign_extend(x, 12) if signed else x

def stype_imm(inst: int) -> int:
    """
    Decodes an S-type immediate value.

    Args:
       inst(int): Word-length instruction in bytes.

    Returns:
       int: S-type immediate value.
    """
    return sign_extend((((inst >> 25) & 0b1111111) << 5) |
                       ((inst >> 7) & 0b11111), 12)

def utype_imm(inst: int) -> int:
    """
    Decodes an U-type immediate value.

    Args:
       inst(int): Word-length instruction in bytes.

    Returns:
       int: U-type immediate value.
    """
    return sign_extend(((inst >> 12) & 0xFFFFFF) << 12)

def sign_extend(val: int, bit_len: int=32) -> int:
    """
    Sign-extends a value to a given extension
    bit length (default = 32 bits).

    Args:
        val(int): Value to extend.
        bit_len(int): Extension length.
    
    Returns:
        int: Sign-extended value.
    """
    sign_bit = 1 << (bit_len - 1)
    return (val & (sign_bit - 1)) - (val & sign_bit)
    

def decode_instruction(inst: int) -> RVInst:
    """
    Decodes a RV32I instruction.

    RISC-V instructions contain their opcode in the lower 7 bits. 
    In the specs, one or more logical instructions are mapped to 
    the same opcode. The funct3 field acts as a selector for R-, 
    I-, S-, and B-type instructions. For R-type instructions, 
    the funct7 field selects instructions with the same opcode 
    and funct3, e.g. add and sub instructions.
        
    Args:
        inst(int): Instruction as an integer.

    Returns:
        RVInst: RISC-V instruction data type.
    """
    opcode = inst & 0b1111111
    funct3 = (inst >> 12) & 0b111
    funct7 = (inst >> 25) & 0b1111111
    rd = (inst >> 7) & 0b11111
    rs1 = (inst >> 15) & 0b11111
    rs2 = (inst >> 20) & 0b11111
    inst_type = mnemonic = None
    imm_signed = True

    if opcode == Opcode.LUI:
        inst_type = UType
        mnemonic = Instruction.LUI
    elif opcode == Opcode.AUIPC:
        inst_type = UType
        mnemonic = Instruction.AUIPC
    elif opcode == Opcode.JAL:
        inst_type = JType
        mnemonic = Instruction.JAL
    elif opcode == Opcode.JALR:
        inst_type = IType
        mnemonic = Instruction.JALR
    elif opcode == Opcode.BRANCH:
        inst_type = BType
        if funct3 == Funct3.BEQ:
            mnemonic = Instruction.BEQ
        elif funct3 == Funct3.BNE:
            mnemonic = Instruction.BNE
        elif funct3 == Funct3.BLT:
            mnemonic = Instruction.BLT
        elif funct3 == Funct3.BGE:
            mnemonic = Instruction.BGE
        elif funct3 == Funct3.BLTU:
            mnemonic = Instruction.BLTU
            imm_signed = False
        elif funct3 == Funct3.BGEU:
            mnemonic = Instruction.BGEU
            imm_signed = False
        else:
            raise DecodeError("Invalid branch instruction!")
    elif opcode == Opcode.LOAD:
        inst_type = IType
        if funct3 == Funct3.LB:
            mnemonic = Instruction.LB
        elif funct3 == Funct3.LH:
            mnemonic = Instruction.LH
        elif funct3 == Funct3.LW:
            mnemonic = Instruction.LW
        elif funct3 == Funct3.LBU:
            mnemonic = Instruction.LBU
            imm_signed = False
        elif funct3 == Funct3.LHU:
            mnemonic = Instruction.LHU
            imm_signed = False
        else:
            raise DecodeError("Invalid load instruction!")
    elif opcode == Opcode.STORE:
        inst_type = SType
        if funct3 == Funct3.SB:
            mnemonic = Instruction.SB
        elif funct3 == Funct3.SH:
            mnemonic = Instruction.SH
        elif funct3 == Funct3.SW:
            mnemonic = Instruction.SW
        else:
            raise DecodeError("Invalid store instruction!")
    elif opcode == Opcode.IMMEDIATE:
        inst_type = IType
        if funct3 == Funct3.ADDI:
            mnemonic = Instruction.ADDI
        elif funct3 == Funct3.SLTI:
            mnemonic = Instruction.SLTI
        elif funct3 == Funct3.SLTIU:
            mnemonic = Instruction.SLTIU
            imm_signed = False
        elif funct3 == Funct3.XORI:
            mnemonic = Instruction.XORI
        elif funct3 == Funct3.ORI:
            mnemonic = Instruction.ORI
        elif funct3 == Funct3.ANDI:
            mnemonic = Instruction.ANDI
        elif funct3 == Funct3.SLLI:
            mnemonic = Instruction.SLLI
        elif funct3 == Funct3.SRLI:
            if funct7 == Funct7.SRLI:
                mnemonic = Instruction.SRLI
            elif funct7 == Funct7.SRAI:
                mnemonic = Instruction.SRAI
            else:
                raise DecodeError("Invalid funct7!")
        else:
            raise DecodeError("Invalid funct3!")
    elif opcode == Opcode.ARITHMETIC:
        inst_type = RType
        if funct3 == Funct3.ADD:
            if funct7 == Funct7.ADD:
                mnemonic = Instruction.ADD
            elif funct7 == Funct7.SUB:
                mnemonic = Instruction.SUB
        elif funct3 == Funct3.SLL:
            mnemonic = Instruction.SLL
        elif funct3 == Funct3.SLT:
            mnemonic = Instruction.SLT
        elif funct3 == Funct3.SLTU:
            mnemonic = Instruction.SLTU
        elif funct3 == Funct3.XOR:
            mnemonic = Instruction.XOR
        elif funct3 == Funct3.SRL:
            if funct7 == Funct7.SRL:
                mnemonic = Instruction.SRL
            elif funct7 == Funct7.SRA:
                mnemonic = Instruction.SRA
        elif funct3 == Funct3.OR:
            mnemonic = Instruction.OR
        elif funct3 == Funct3.AND:
            mnemonic = Instruction.AND
        else:
            raise DecodeError("Invalid arithmetic instruction!")
    elif opcode == Opcode.FENCES:
        inst_type = IType
        if funct3 == Funct3.FENCE:
            mnemonic = Instruction.FENCE
        elif funct3 == Funct3.FENCE_I:
            mnemonic = Instruction.FENCE.I
        else:
            raise DecodeError("Invalid fence instruction!")
    elif opcode == Opcode.SYSTEM:
        # System instructions also differ from the
        # pre-defined types.
        inst_type = IType
        if funct3 == Funct3.ECALL:
            # ECALL and EBREAK have the same funct3.
            # Bits [20:31] are used to differentiate them,
            # as 0 == ECALL and 1 == EBREAK. We can get
            # this from rs2 calculated earlier
            if rs2 == 0:
                mnemonic = Instruction.ECALL
            elif rs2 == 1:
                mnemonic = Instruction.EBREAK
            else:
                raise DecodeError("Invalid environment instruction!")
        elif funct3 == Funct3.CSRRW:
            mnemonic = Instruction.CSRRW
        elif funct3 == Funct3.CSRRS:
            mnemonic = Instruction.CSRRS
        elif funct3 == Funct3.CSRRC:
            mnemonic = Instruction.CSRRC
        elif funct3 == Funct3.CSRRWI:
            mnemonic = Instruction.CSRRWI
        elif funct3 == Funct3.CSRRSI:
            mnemonic = Instruction.CSRRSI
        elif funct3 == Funct3.CSRRCI:
            mnemonic = Instruction.CSRRCI
        else:
            raise DecodeError("Invalid system instruction!")        
    else:
        raise DecodeError("Invalid opcode!")

    if inst_type == RType:
        return RType(mnemonic=mnemonic, opcode=opcode, rd=rd,
                     funct3=funct3, rs1=rs1, rs2=rs2, funct7=funct7)
    elif inst_type == IType:
        return IType(mnemonic=mnemonic, opcode=opcode,
                     imm=itype_imm(inst, signed=imm_signed),
                     rd=rd, funct3=funct3, rs1=rs1)
    elif inst_type == SType:
        return SType(mnemonic=mnemonic, opcode=opcode, imm=stype_imm(inst),
                     funct3=funct3, rs1=rs1, rs2=rs2)
    elif inst_type == BType:
        return BType(mnemonic=mnemonic, opcode=opcode,
                     imm=btype_imm(inst, signed=imm_signed),
                     funct3=funct3, rs1=rs1, rs2=rs2)
    elif inst_type == UType:
        return UType(mnemonic=mnemonic, opcode=opcode, imm=utype_imm(inst), rd=rd)
    elif inst_type == JType:
        return JType(mnemonic=mnemonic, opcode=opcode, imm=jtype_imm(inst), rd=rd)
    else:
        raise DecodeError("A serious error occurred.")



import pytest

from voyagercpu.utils import *

def test_abi_register_name_dict():
    #See: https://riscv.org/wp-content/uploads/2017/05/riscv-spec-v2.2.pdf
    d = abi_register_name_dict()
    assert d["x0"] == "zero"
    assert d["x1"] == "ra"
    assert d["x4"] == "tp"
    assert d["x7"] == "t2"
    assert d["x10"] == "a0"
    assert d["x13"] == "a3"
    assert d["x17"] == "a7"
    assert d["x18"] == "s2"
    assert d["x27"] == "s11"
    assert d["x28"] == "t3"
    assert d["x31"] == "t6"
    assert d["pc"] == "pc"



import struct

from decoder import *
from utils import register_names, abi_register_name_dict, logger

# Instructions are always 4-byte aligned for RV32I
INST_ALIGN = 4
PC_REG_INDEX = 32

class AlignmentError(Exception):
    pass

class VoyagerCPU:
    def __init__(self, start_pc=0):
        self.regfile = self.reset_regs()
        self.regfile[PC_REG_INDEX] = start_pc
        self.time_period = 0

    def __str__(self) -> str:
        dump_str = f"Time period: {self.time_period}\n"
        dump_str += "Register states:\n"
        for i, r in enumerate(self.regfile.keys()):
            dump_str += f"{REG_DICT[i]}: {self.regfile[r]} \t"
            if (i + 1) % 5 == 0:
                dump_str += "\n"
        return dump_str

    def reset_regs(self) -> dict:
        return { i: 0 for i, _ in enumerate(register_names()) }

    def __fetch(self, ram):
        raw_inst = ram.read(self.regfile[PC_REG_INDEX],
                            INST_ALIGN)
        logger.debug(f"Fetched instruction: 0x{raw_inst.hex()}")
        # Convert to little endian
        raw_inst_bin = struct.unpack("<I", raw_inst)[0]
        return raw_inst_bin

    def __decode(self, raw_inst: int) -> RVInst:
        decoded_inst = decode_instruction(raw_inst)
        logger.debug(f"Decoded: {decoded_inst}")
        return decoded_inst

    def __execute(self, inst: RVInst):
        mne = inst.mnemonic
        r = self.regfile # For brevity

        if type(inst) == UType:
            if mne == Instruction.LUI:
                r[inst.rd] = inst.imm
            elif mne == Instruction.AUIPC:
                r[inst.rd] = r[PC_REG_INDEX] + inst.imm
        elif type(inst) == JType:
            if mne == Instruction.JAL:
                r[inst.rd] = r[PC_REG_INDEX] + INST_ALIGN
                r[PC_REG_INDEX] += inst.imm
        elif type(inst) == BType:
            if mne == Instruction.BEQ and inst.rs1 == inst.rs2:
                r[PC_REG_INDEX] += inst.imm
            elif mne == Instruction.BNE and inst.rs1 != inst.rs2:
                r[PC_REG_INDEX] += inst.imm
            elif (mne == Instruction.BLT or mne == Instruction.BLTU) \
                 and inst.rs1 < inst.rs2:
                r[PC_REG_INDEX] += inst.imm
            elif (mne == Instruction.BGE or mne == Instruction.BGEU) \
                 and inst.rs1 >= inst.rs2:
                r[PC_REG_INDEX] += inst.imm
        elif type(inst) == SType:
            if mne == Instruction.SB:
                ram[inst.rs1 + inst.imm] = struct.unpack("B", r[inst.rs2] & 0xFF)
            elif mne == Instruction.SH:
                ram[inst.rs1 + inst.imm] = struct.unpack("H", r[inst.rs2] & 0xFFFF)
            elif mne == Instruction.SW:
                ram[inst.rs1 + inst.imm] = struct.unpack("I", r[inst.rs2] & 0xFFFFFFFF)
        elif type(inst) == IType:
            # JALR
            if mne == Instruction.JALR:
                r[inst.rd] = r[PC_REG_INDEX] + INST_ALIGN
                r[PC_REG_INDEX] = r[inst.rs1] + inst.imm
            # Load instructions
            elif mne == Instruction.LB or mne == Instruction.LBU:
                r[inst.rd] = ram[inst.rs1 + inst.imm]
            elif mne == Instruction.LH or mne == Instruction.LHU:
                addr = inst.rs1 + inst.imm
                r[inst.rd] = ram[addr:addr+2]
            elif mne == Instruction.LW:
                addr = inst.rs1 + inst.imm
                r[inst.rd] = ram[addr:addr+4]
            # Arithmetic immediate instructions
            elif mne == Instruction.ADDI:
                r[inst.rd] = r[inst.rs1] + inst.imm
            elif mne == Instruction.SLTI or mne == Instruction.SLTIU:
                r[inst.rd] = 1 if r[inst.rs1] < inst.imm else 0
            elif mne == Instruction.XORI:
                r[inst.rd] = r[inst.rs1] ^ inst.imm
            elif mne == Instruction.ORI:
                r[inst.rd] = r[inst.rs1] | inst.imm
            elif mne == Instruction.ANDI:
                r[inst.rd] = r[inst.rs1] & inst.imm
            elif mne == Instruction.SLLI:
                r[inst.rd] = r[inst.rs1] << inst.imm
            elif mne == Instruction.SRLI:
                sign_bit = r[inst.rs1] >> 31
                res = r[inst.rs1] >> (inst.imm & 0x1F)
                res |= (0xFFFFFFFF * sign_bit) << (32 - (inst.imm & 0x1F))
                r[inst.rd] = res
            elif mne == Instruction.SRAI:
                r[inst.rd] = r[inst.rs1] >> inst.imm
        elif type(inst) == RType:
            if mne == Instruction.ADD:
                r[inst.rd] = r[inst.rs1] + r[inst.rs2]
            elif mne == Instruction.SUB:
                r[inst.rd] = r[inst.rs1] - r[inst.rs2]
            elif mne == Instruction.SLL:
                r[inst.rd] = r[inst.rs1] << r[inst.rs2]
            elif mne == Instruction.SLT:
                r[inst.rd] = 1 if r[inst.rs1] < r[inst.rs2] else 0
            elif mne == Instruction.SLTU:
                rs1_u = r[inst.rs1] & 0xFFFFFFFF
                rs2_u = r[inst.rs2] & 0xFFFFFFFF
                r[inst.rd] = 1 if rs1_u < rs2_u else 0
            elif mne == Instruction.XOR:
                r[inst.rd] = r[inst.rs1] ^ r[inst.rs2]
            elif mne == Instruction.SRL:
                sign_bit = r[inst.rs1] >> 31
                res = r[inst.rs1] >> (r[inst.rs2] & 0x1F)
                res |= (0xFFFFFFFF * sign_bit) << (32 - (r[inst.rs2] & 0x1F))
                r[inst.rd] = res
            elif mne == Instruction.SRA:
                r[inst.rd] = r[inst.rs1] >> r[inst.rs2]
            elif mne == Instruction.OR:
                r[inst.rd] = r[inst.rs1] | r[inst.rs2]
            elif mne == Instruction.AND:
                r[inst.rd] = r[inst.rs1] & r[inst.rs2]
        else:
            logger.debug("System instructions not implemented")

    def next_cycle(self, ram):
        raw_inst = self.__fetch(ram)
        decoded_inst = self.__decode(raw_inst)
        self.__execute(decoded_inst)

        # Confirm that the PC is aligned at a multiple of 4
        if self.regfile[PC_REG_INDEX] & 0b11:
            raise AlignmentError(f"Program counter is misaligned! - " \
                                 f"PC: {self.regfile[PC_REG_INDEX]}")
        
        self.regfile[PC_REG_INDEX] += INST_ALIGN
        logger.debug(self.__str__())
        self.time_period += 1



class VoyagerRAM:
    DEFAULT_RAM_SIZE = 0x1000

    def __init__(self, ram_size=DEFAULT_RAM_SIZE):
        self.ram_size = ram_size
        self.ram = b"\x00" * ram_size

    def __str__(self):
        ram_str = ""
        for i, data in enumerate(self.ram):
            if i % 8 == 0:
                ram_str += "\n"
            ram_str += f"0x{i:03X}: {data:02X}  "
        return ram_str

    def write(self, data, addr=0):
        self.ram = self.ram[:addr] + data + self.ram[addr+len(data):]

    def read(self, start_idx, end_idx=1):
        return self.ram[start_idx:start_idx+end_idx]

    def dump(self):
        print(self.__str__())



import pytest

from voyagercpu.decoder import *

def test_decode_instruction_type():
    # jal x0, 72
    inst = decode_instruction(0x0480006f)
    assert type(inst) == JType
    assert inst.mnemonic == Instruction.JAL
    assert inst.imm == 72
    assert REG_DICT[inst.rd] == "x0"

    # jal x0, -8
    inst = decode_instruction(0xff9ff06f)
    assert type(inst) == JType
    assert inst.mnemonic == Instruction.JAL
    assert inst.imm == -8
    assert REG_DICT[inst.rd] == "x0"

    # addi x8, x0, 0
    inst = decode_instruction(0x00000413)
    assert type(inst) == IType
    assert inst.mnemonic == Instruction.ADDI
    assert inst.imm == 0
    assert REG_DICT[inst.rd] == "x8"
    assert REG_DICT[inst.rs1] == "x0"

    # addi x12, x0, 0
    inst = decode_instruction(0x00000613)
    assert type(inst) == IType
    assert inst.mnemonic == Instruction.ADDI
    assert inst.imm == 0
    assert REG_DICT[inst.rd] == "x12"
    assert REG_DICT[inst.rs1] == "x0"

    # addi x15, x0, 0
    inst = decode_instruction(0x00000793)
    assert type(inst) == IType
    assert inst.mnemonic == Instruction.ADDI
    assert inst.imm == 0
    assert REG_DICT[inst.rd] == "x15"
    assert REG_DICT[inst.rs1] == "x0"    

    # auipc x30, 4096
    inst = decode_instruction(0x00001f17)
    assert type(inst) == UType
    assert inst.mnemonic == Instruction.AUIPC
    assert inst.imm == 4096
    assert REG_DICT[inst.rd] == "x30"

    # auipc x10, -8192
    inst = decode_instruction(0xffffe517)
    assert type(inst) == UType
    assert inst.mnemonic == Instruction.AUIPC
    assert inst.imm == -8192
    assert REG_DICT[inst.rd] == "x10"

    # lui x7, 267390976
    inst = decode_instruction(0x0ff013b7)
    assert type(inst) == UType
    assert inst.mnemonic == Instruction.LUI
    assert inst.imm == 267390976
    assert REG_DICT[inst.rd] == "x7"

    # sw x3,-60(x30)
    inst = decode_instruction(0xfc3f2223)
    assert type(inst) == SType
    assert inst.mnemonic == Instruction.SW
    assert inst.imm == -60
    assert REG_DICT[inst.rs1] == "x30"
    assert REG_DICT[inst.rs2] == "x3"

    # sw x1,128(x21)
    inst = decode_instruction(0x081aa023)
    assert type(inst) == SType
    assert inst.mnemonic == Instruction.SW
    assert inst.imm == 128
    assert REG_DICT[inst.rs1] == "x21"
    assert REG_DICT[inst.rs2] == "x1"

    # csrrs x30,mcause,x0
    inst = decode_instruction(0x34202f73)
    assert type(inst) == IType
    assert inst.mnemonic == Instruction.CSRRS
    assert REG_DICT[inst.rd] == "x30"
    assert REG_DICT[inst.rs1] == "x0"

    # ori x3, x3, 1337
    inst = decode_instruction(0x5391e193)
    assert type(inst) == IType
    assert inst.mnemonic == Instruction.ORI
    assert inst.imm == 1337
    assert REG_DICT[inst.rd] == "x3"
    assert REG_DICT[inst.rs1] == "x3"

    # xori x5, x7, -247
    inst = decode_instruction(0xf093c293)
    assert type(inst) == IType
    assert inst.mnemonic == Instruction.XORI
    assert inst.imm == -247
    assert REG_DICT[inst.rd] == "x5"
    assert REG_DICT[inst.rs1] == "x7"

    # bne x10,x0, 0
    inst = decode_instruction(0x00051063)
    assert type(inst) == BType
    assert inst.mnemonic == Instruction.BNE
    assert inst.imm == 0
    assert REG_DICT[inst.rs1] == "x10"
    assert REG_DICT[inst.rs2] == "x0"

    # beq x0, x3, -32
    inst = decode_instruction(0xfe3000e3)
    assert type(inst) == BType
    assert inst.mnemonic == Instruction.BEQ
    assert inst.imm == -32
    assert REG_DICT[inst.rs1] == "x0"
    assert REG_DICT[inst.rs2] == "x3"

    # ecall
    inst = decode_instruction(0x00000073)
    assert type(inst) == IType
    assert inst.mnemonic == Instruction.ECALL

    # sub x2, x4, x5
    inst = decode_instruction(0x40520133)
    assert type(inst) == RType
    assert inst.mnemonic == Instruction.SUB
    assert REG_DICT[inst.rd] == "x2"
    assert REG_DICT[inst.rs1] == "x4"
    assert REG_DICT[inst.rs2] == "x5"



#!/usr/bin/env python3
import struct

from ram import VoyagerRAM
from cpu import VoyagerCPU
from utils import logger

from elftools.elf.elffile import ELFFile

REPL_PROMPT = "> Next step (n), view registers (r), memory (m), quit (q), or enter N steps: "
PROGRAM_PROMPT = "> Select test program - (1, default) rv32ui-p-xor, " \
    "(2) rv32ui-p-add, " \
    "(3) rv32ui-p-srai: "
SEG_N = 1
TEST_PROGRAM_PATH =  "./tests/riscv-tests-prebuilt-binaries/isa/rv32ui/"

if __name__ == "__main__":
    voyager_cpu = VoyagerCPU()
    voyager_ram = VoyagerRAM()

    usr_in = input(PROGRAM_PROMPT)
    f = TEST_PROGRAM_PATH
    
    if "2" in usr_in:
        f += "rv32ui-p-add"
    elif "3" in usr_in:
        f += "rv32ui-p-srai"
    else:
        f += "rv32ui-p-xor"

    print(f"Loading {f}...")

    with open(f, "rb") as ff:
        e = ELFFile(ff)
        for i, s in enumerate(e.iter_segments()):
            print(f"Segment {i} type: {s['p_type']}")

        print(f"Loading segment {SEG_N}...")
        seg = e.get_segment(SEG_N)
        voyager_ram.write(seg.data())

    while True:
        usr_in = input(REPL_PROMPT)
        if "r" in usr_in:
            print(sphinx_cpu)
        elif "m" in usr_in:
            print(sphinx_ram)
        elif "q" in usr_in:
            break
        elif "n" in usr_in:
            voyager_cpu.next_cycle(sphinx_ram)
            logger.debug(f"Step: {i}")
        elif usr_in.isdigit():
            for i in range(int(usr_in)):
                if i % 5 == 0:
                    print(f"Step: {i}")
                voyager_cpu.next_cycle(sphinx_ram)
        else:
            print("Try again")
    print("Bye!")



import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

def abi_register_names() -> list:
    """
    Returns a list of RISC-V ABI register names.

    Returns:
        list: List of ABI register names as strings.
    """
    return ["zero", "ra", "sp", "gp", "tp"] + \
        [f"t{i}" for i in range(3)] + \
        ["s0", "s1"] + \
        [f"a{i}" for i in range(8)] + \
        [f"s{i}" for i in range(2, 12)] + \
        [f"t{i}" for i in range(3, 7)] + ["pc"]

def register_names() -> list:
    """
    Returns a list of basic RISC-V register names.

    These have the form x0-31. Their ABI names 
    may be derived using `abi_register_names()`.

    Returns:
       list: List of register names as strings.
    """
    return [f"x{i}" for i in range(32)] + ["pc"]

def abi_register_name_dict() -> dict:
    """
    Creates a dict of standard registers and their ABI names, 
    e.g. ("x0" -> "zero"), ("x2" -> "sp") etc.
    """
    return { reg: abi for reg, abi in
             zip(register_names(), abi_register_names()) }

def int_to_hex_str(val: int,
                   endian: str="little",
                   prefix: str="0x") -> str:
    # Get the hex string of val and pad it with a leading
    # zero if its length isn't even
    val_hex_str = hex(val)[2:]
    if val_hex_str % 2 != 0:
        val_hex_str = "0" + val_hex_str

    if endian == "big":
        return prefix + val_hex_str
    elif endian == "little":
        return prefix + "".join(
            reversed([a[i:i+2] for i in range(0, len(a), 2)]))
    else:
        raise ValueError("Endian must be \"little\" or \"big\"!")
